@@ -1,12 +1,12 @@
 from __future__ import annotations
 
 from itertools import product
-from typing import Any, Dict, Iterator, Optional, Sequence, Tuple, Union
+from typing import Any, Dict, Iterator, Optional, Sequence, Tuple, Union, no_type_check
 from uuid import UUID, uuid4
 from warnings import warn
 
 import numpy as np
-from pydantic import Field, root_validator, validator
+from pydantic import Field, PrivateAttr, root_validator, validator
 
 from ._base_model import UseqModel
 from ._channel import Channel
@@ -21,32 +21,106 @@ POSITION = "p"
 Z = "z"
 INDICES = (TIME, POSITION, CHANNEL, Z)
 
+Undefined = object()
+
 
 class MDASequence(UseqModel):
+    """A sequence of MDA (Multi-Dimensional Acquisition) events.
+
+    This is the core of the useq library, and is used define a sequence of events to be
+    run on a microscope. It object may be constructed manually, or from file (e.g. json
+    or yaml).
+
+    The object itself acts as an iterator for `useq.MDAEvent` objects:
+
+    Parameters
+    ----------
+    axis_order : str
+        The order of the axes in the sequence. Must be a permutation of "tpcz". The
+        default is "tpcz".
+    stage_positions : Tuple[Position, ...]
+        The stage positions to visit. (each with `x`, `y`, `z`, `name`, and `z_plan`,
+        all of which are optional).
+    channels : Tuple[Channel, ...]
+        The channels to acquire. see `Channel`.
+    time_plan : AnyTimePlan
+        The time plan to follow. One of `TIntervalDuration`, `TIntervalLoops`,
+        `TDurationLoops`, `MultiPhaseTimePlan`, or `NoT`
+    z_plan : AnyZPlan
+        The z plan to follow. One of `ZTopBottom`, `ZRangeAround`, `ZAboveBelow`,
+        `ZRelativePositions`, `ZAbsolutePositions`, or `NoZ`.
+
+    Examples
+    --------
+    >>> from useq import MDASequence, Position, Channel, TIntervalDuration
+    >>> seq = MDASequence(
+    ...     time_plan={"interval": 0.1, "loops": 2},
+    ...     stage_positions=[(1, 1, 1)],
+    ...     z_plan={"range": 3, "step": 1},
+    ...     channels=[{"config": "DAPI", "exposure": 1}]
+    ... )
+    >>> print(seq)
+    Multi-Dimensional Acquisition ▶ nt: 2, np: 1, nc: 1, nz: 4
+
+    >>> for event in seq:
+    ...     print(event)
+
+    >>> print(seq.yaml())
+    channels:
+    - config: DAPI
+      exposure: 1.0
+    stage_positions:
+    - x: 1.0
+      y: 1.0
+      z: 1.0
+    time_plan:
+      interval: '0:00:00.100000'
+      loops: 2
+    z_plan:
+      range: 3.0
+      step: 1.0
+    """
+
     metadata: Dict[str, Any] = Field(default_factory=dict)
     axis_order: str = "".join(INDICES)
     stage_positions: Tuple[Position, ...] = Field(default_factory=tuple)
     channels: Tuple[Channel, ...] = Field(default_factory=tuple)
     time_plan: AnyTimePlan = Field(default_factory=NoT)
     z_plan: AnyZPlan = Field(default_factory=NoZ)
-    uid: UUID = Field(default_factory=uuid4)
+
+    _uid: UUID = PrivateAttr(default_factory=uuid4)
+    _length: Optional[int] = PrivateAttr(default=None)
+
+    @property
+    def uid(self) -> UUID:
+        """A unique identifier for this sequence."""
+        return self._uid
+
+    @no_type_check
+    def replace(
+        self,
+        metadata: Dict[str, Any] = Undefined,
+        axis_order: str = Undefined,
+        stage_positions: Tuple[Position, ...] = Undefined,
+        channels: Tuple[Channel, ...] = Undefined,
+        time_plan: AnyTimePlan = Undefined,
+        z_plan: AnyZPlan = Undefined,
+    ) -> MDASequence:
+        """Return a new `MDAsequence` replacing specified fields with new values."""
+        kwargs = {k: v for k, v in locals().items() if v is not Undefined}
+        state = self.dict(exclude={"uid"})
+        return type(self)(**{**state, **kwargs})
 
     def __hash__(self) -> int:
         return hash(self.uid)
 
     @validator("z_plan", pre=True)
     def validate_zplan(cls, v: Any) -> Union[dict, NoZ]:
-        if not v:
-            return NoZ()
-        return v
+        return v or NoZ()
 
     @validator("time_plan", pre=True)
     def validate_time_plan(cls, v: Any) -> Union[dict, NoT]:
-        if isinstance(v, (tuple, list)):
-            return {"phases": v}
-        if not v:
-            return NoT()
-        return v
+        return {"phases": v} if isinstance(v, (tuple, list)) else v or NoT()
 
     @validator("stage_positions", pre=True)
     def validate_positions(cls, v: Any) -> list:
@@ -90,9 +164,8 @@ class MDASequence(UseqModel):
         else:
             return False
 
-    @classmethod
+    @staticmethod
     def _check_order(
-        cls,
         order: str,
         z_plan: AnyZPlan = None,
         stage_positions: Sequence[Position] = (),
@@ -123,41 +196,29 @@ class MDASequence(UseqModel):
 
         return order
 
-    def add_channel(
-        self,
-        config: str,
-        group: str = "Channel",
-        exposure: int = None,
-        do_stack: bool = False,
-    ) -> None:
-        new = Channel(config=config, group=group, exposure=exposure, do_stack=do_stack)
-        self.channels = tuple(self.channels) + (new,)
-
-    def remove_channel(self, **kwargs: Union[str, int, bool]) -> None:
-        to_pop = [
-            c
-            for c in self.channels
-            if any(getattr(c, k) == v for k, v in kwargs.items())
-        ]
-        self.channels = tuple(i for i in self.channels if i not in to_pop)
-
     def __str__(self) -> str:
-        out = "Multi-Dimensional Acquisition ▶ "
         shape = [
             f"n{k.lower()}: {len(list(self.iter_axis(k)))}" for k in self.axis_order
         ]
-        out += ", ".join(shape)
-        return out
+        return "Multi-Dimensional Acquisition ▶ " + ", ".join(shape)
 
-    # def __len__(self):
-    #     # np.prod(self.shape)
-    #     return len(list(self.iter_events()))
+    def __len__(self) -> int:
+        if self._length is None:
+            self._length = len(list(self.iter_events()))
+        return self._length
 
     @property
     def shape(self) -> Tuple[int, ...]:
-        # NOTE: Doesn't account for skipped Z or channel frames
-        shp = (len(list(self.iter_axis(k))) for k in self.axis_order)
-        return tuple(s for s in shp if s)
+        # NOTE: Doesn't account for jagged arrays, like skipped Z or channel frames
+        return tuple(s for s in self.sizes.values() if s)
+
+    @property
+    def sizes(self) -> Dict[str, int]:
+        return {k: len(list(self.iter_axis(k))) for k in self.axis_order}
+
+    @property
+    def used_axes(self) -> str:
+        return "".join(k for k in self.axis_order if self.sizes[k])
 
     def iter_axis(self, axis: str) -> Iterator[Union[Position, Channel, float]]:
         yield from {
@@ -168,18 +229,25 @@ class MDASequence(UseqModel):
         }[axis]
 
     def __iter__(self) -> Iterator[MDAEvent]:  # type: ignore
-        yield from self.iter_events(self.axis_order)
+        yield from self.iter_events()
 
     class _SkipFrame(Exception):
         pass
 
-    def iter_events(self, order: str = None) -> Iterator[MDAEvent]:
+    def iter_events(self) -> Iterator[MDAEvent]:
+        """Iterate over all events in the MDA sequence.
 
-        order = self._check_order(order or self.axis_order)
-        # strip dimensions that have no length
-        order = "".join(i for i in order if list(self.iter_axis(i)))
+        This does the job of iterating over all the frames in the MDA sequence,
+        handling the merging of z plans in channels and stage positions.
 
-        for item in product(*(enumerate(self.iter_axis(ax)) for ax in order)):
+        Yields
+        ------
+        Iterator[MDAEvent]
+        """
+        order = self.used_axes
+
+        event_iterator = (enumerate(self.iter_axis(ax)) for ax in order)
+        for global_index, item in enumerate(product(*event_iterator)):
             if not item:  # the case with no events
                 continue
 
@@ -198,6 +266,8 @@ class MDASequence(UseqModel):
                 z_pos = (
                     self._combine_z(_ev[Z][1], index[Z], channel, position)
                     if Z in _ev
+                    else position.z
+                    if position
                     else None
                 )
             except self._SkipFrame:
@@ -209,12 +279,14 @@ class MDASequence(UseqModel):
             yield MDAEvent(
                 index=index,
                 min_start_time=time,
+                pos_name=getattr(position, "name", None),
                 x_pos=getattr(position, "x", None),
                 y_pos=getattr(position, "y", None),
                 z_pos=z_pos,
                 exposure=getattr(channel, "exposure", None),
                 channel=_channel,
                 sequence=self,
+                global_index=global_index,
             )
 
     def _combine_z(
